]> git.sesse.net Git - nageru/blobdiff - mixer.cpp
Add some GUI elements (hooked up) to help tuning the compressor.
[nageru] / mixer.cpp
index 7584751b9725c15b4b3a7202d39ea403d343c7a4..861ea94d221bae78c93c06ec41f203aa30cd0905 100644 (file)
--- a/mixer.cpp
+++ b/mixer.cpp
@@ -70,7 +70,9 @@ Mixer::Mixer(const QSurfaceFormat &format, unsigned num_cards)
          num_cards(num_cards),
          mixer_surface(create_surface(format)),
          h264_encoder_surface(create_surface(format)),
-         level_compressor(OUTPUT_FREQUENCY)
+         level_compressor(OUTPUT_FREQUENCY),
+         limiter(OUTPUT_FREQUENCY),
+         compressor(OUTPUT_FREQUENCY)
 {
        httpd.start(9095);
 
@@ -122,7 +124,7 @@ Mixer::Mixer(const QSurfaceFormat &format, unsigned num_cards)
                        [this]{
                                resource_pool->clean_context();
                        });
-               card->resampler.reset(new Resampler(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
+               card->resampling_queue.reset(new ResamplingQueue(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
                card->usb->configure_card();
        }
 
@@ -151,6 +153,10 @@ Mixer::Mixer(const QSurfaceFormat &format, unsigned num_cards)
        r128.integr_start();
 
        locut.init(FILTER_HPF, 2);
+
+       // hlen=16 is pretty low quality, but we use quite a bit of CPU otherwise,
+       // and there's a limit to how important the peak meter is.
+       peak_resampler.setup(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY * 4, /*num_channels=*/2, /*hlen=*/16);
 }
 
 Mixer::~Mixer()
@@ -180,10 +186,10 @@ int unwrap_timecode(uint16_t current_wrapped, int last)
        }
 }
 
-float find_peak(const vector<float> &samples)
+float find_peak(const float *samples, size_t num_samples)
 {
        float m = fabs(samples[0]);
-       for (size_t i = 1; i < samples.size(); ++i) {
+       for (size_t i = 1; i < num_samples; ++i) {
                m = std::max(m, fabs(samples[i]));
        }
        return m;
@@ -247,7 +253,7 @@ void Mixer::bm_frame(unsigned card_index, uint16_t timecode,
                if (dropped_frames > FPS * 2) {
                        fprintf(stderr, "Card %d lost more than two seconds (or time code jumping around), resetting resampler\n",
                                card_index);
-                       card->resampler.reset(new Resampler(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
+                       card->resampling_queue.reset(new ResamplingQueue(OUTPUT_FREQUENCY, OUTPUT_FREQUENCY, 2));
                } else if (dropped_frames > 0) {
                        // Insert silence as needed.
                        fprintf(stderr, "Card %d dropped %d frame(s) (before timecode 0x%04x), inserting silence.\n",
@@ -255,10 +261,10 @@ void Mixer::bm_frame(unsigned card_index, uint16_t timecode,
                        vector<float> silence;
                        silence.resize((OUTPUT_FREQUENCY / FPS) * 2);
                        for (int i = 0; i < dropped_frames; ++i) {
-                               card->resampler->add_input_samples((unwrapped_timecode - dropped_frames + i) / double(FPS), silence.data(), (OUTPUT_FREQUENCY / FPS));
+                               card->resampling_queue->add_input_samples((unwrapped_timecode - dropped_frames + i) / double(FPS), silence.data(), (OUTPUT_FREQUENCY / FPS));
                        }
                }
-               card->resampler->add_input_samples(unwrapped_timecode / double(FPS), audio.data(), num_samples);
+               card->resampling_queue->add_input_samples(unwrapped_timecode / double(FPS), audio.data(), num_samples);
        }
 
        // Done with the audio, so release it.
@@ -464,7 +470,7 @@ void Mixer::thread_func()
                for (unsigned card_index = 0; card_index < num_cards; ++card_index) {
                        input_frames.push_back(bmusb_current_rendering_frame[card_index]);
                }
-               const int64_t av_delay = TIMEBASE / 10;  // Corresponds to the fixed delay in resampler.h. TODO: Make less hard-coded.
+               const int64_t av_delay = TIMEBASE / 10;  // Corresponds to the fixed delay in resampling_queue.h. TODO: Make less hard-coded.
                h264_encoder->end_frame(fence, pts_int + av_delay, input_frames);
                ++frame;
                pts_int += TIMEBASE / FPS;
@@ -531,7 +537,7 @@ void Mixer::process_audio_one_frame()
                samples_card.resize((OUTPUT_FREQUENCY / FPS) * 2);
                {
                        unique_lock<mutex> lock(cards[card_index].audio_mutex);
-                       if (!cards[card_index].resampler->get_output_samples(pts(), &samples_card[0], OUTPUT_FREQUENCY / FPS)) {
+                       if (!cards[card_index].resampling_queue->get_output_samples(pts(), &samples_card[0], OUTPUT_FREQUENCY / FPS)) {
                                printf("Card %d reported previous underrun.\n", card_index);
                        }
                }
@@ -544,28 +550,24 @@ void Mixer::process_audio_one_frame()
        // Cut away everything under 150 Hz; we don't need it for voice,
        // and it will reduce headroom and confuse the compressor.
        // (In particular, any hums at 50 or 60 Hz should be dampened.)
-       locut.render(samples_out.data(), samples_out.size() / 2, 150.0 * 2.0 * M_PI / OUTPUT_FREQUENCY, 0.5f);
+       locut.render(samples_out.data(), samples_out.size() / 2, locut_cutoff_hz * 2.0 * M_PI / OUTPUT_FREQUENCY, 0.5f);
 
        // Apply a level compressor to get the general level right.
        // Basically, if it's over about -40 dBFS, we squeeze it down to that level
        // (or more precisely, near it, since we don't use infinite ratio),
// then apply a makeup gain to get it to -12 dBFS. -12 dBFS is, of course,
        // entirely arbitrary, but from practical tests with speech, it seems to
        // put ut around -23 LUFS, so it's a reasonable starting point for later use.
-       //
-       // TODO: Hook this up to a UI, so we can see the effects, and/or turn it off
-       // to control the gain manually instead. For now, there's only the #if-ed out
-       // code below.
-       //
-       // TODO: Add the actual compressors/limiters (for taking care of transients)
-       // later in the chain.
-       float threshold = 0.01f;   // -40 dBFS.
-       float ratio = 20.0f;
-       float attack_time = 0.1f;
-       float release_time = 10.0f;
-       float makeup_gain = pow(10.0f, 28.0f / 20.0f);  // +28 dB takes us to -12 dBFS.
-       level_compressor.process(samples_out.data(), samples_out.size() / 2, threshold, ratio, attack_time, release_time, makeup_gain);
-       last_gain_staging_db = 20.0 * log10(level_compressor.get_attenuation() * makeup_gain);
+       float ref_level_dbfs = -14.0f;
+       {
+               float threshold = 0.01f;   // -40 dBFS.
+               float ratio = 20.0f;
+               float attack_time = 0.5f;
+               float release_time = 20.0f;
+               float makeup_gain = pow(10.0f, (ref_level_dbfs - (-40.0f)) / 20.0f);  // +26 dB.
+               level_compressor.process(samples_out.data(), samples_out.size() / 2, threshold, ratio, attack_time, release_time, makeup_gain);
+               last_gain_staging_db = 20.0 * log10(level_compressor.get_attenuation() * makeup_gain);
+       }
 
 #if 0
        printf("level=%f (%+5.2f dBFS) attenuation=%f (%+5.2f dB) end_result=%+5.2f dB\n",
@@ -574,8 +576,49 @@ void Mixer::process_audio_one_frame()
                20.0 * log10(level_compressor.get_level() * level_compressor.get_attenuation() * makeup_gain));
 #endif
 
-       // Find peak and R128 levels.
-       peak = std::max(peak, find_peak(samples_out));
+//     float limiter_att, compressor_att;
+
+       // Then a limiter at +0 dB (so, -14 dBFS) to take out the worst peaks only.
+       // Note that since ratio is not infinite, we could go slightly higher than this.
+       // Probably more tuning is warranted here.
+       if (limiter_enabled) {
+               float threshold = pow(10.0f, limiter_threshold_dbfs / 20.0f);
+               float ratio = 30.0f;
+               float attack_time = 0.0f;  // Instant.
+               float release_time = 0.005f;
+               float makeup_gain = 1.0f;  // 0 dB.
+               limiter.process(samples_out.data(), samples_out.size() / 2, threshold, ratio, attack_time, release_time, makeup_gain);
+//             limiter_att = limiter.get_attenuation();
+       }
+
+       // Finally, the real compressor.
+       if (compressor_enabled) {
+               float threshold = pow(10.0f, compressor_threshold_dbfs / 20.0f);
+               float ratio = 20.0f;
+               float attack_time = 0.005f;
+               float release_time = 0.040f;
+               float makeup_gain = 2.0f;  // +6 dB.
+               compressor.process(samples_out.data(), samples_out.size() / 2, threshold, ratio, attack_time, release_time, makeup_gain);
+//             compressor_att = compressor.get_attenuation();
+       }
+
+//     printf("limiter=%+5.1f  compressor=%+5.1f\n", 20.0*log10(limiter_att), 20.0*log10(compressor_att));
+
+       // Upsample 4x to find interpolated peak.
+       peak_resampler.inp_data = samples_out.data();
+       peak_resampler.inp_count = samples_out.size() / 2;
+
+       vector<float> interpolated_samples_out;
+       interpolated_samples_out.resize(samples_out.size());
+       while (peak_resampler.inp_count > 0) {  // About four iterations.
+               peak_resampler.out_data = &interpolated_samples_out[0];
+               peak_resampler.out_count = interpolated_samples_out.size() / 2;
+               peak_resampler.process();
+               size_t out_stereo_samples = interpolated_samples_out.size() / 2 - peak_resampler.out_count;
+               peak = max<float>(peak, find_peak(interpolated_samples_out.data(), out_stereo_samples * 2));
+       }
+
+       // Find R128 levels.
        vector<float> left, right;
        deinterleave_samples(samples_out, &left, &right);
        float *ptrs[] = { left.data(), right.data() };
@@ -670,6 +713,14 @@ void Mixer::channel_clicked(int preview_num)
        theme->channel_clicked(preview_num);
 }
 
+void Mixer::reset_meters()
+{
+       peak_resampler.reset();
+       peak = 0.0f;
+       r128.reset();
+       r128.integr_start();
+}
+
 Mixer::OutputChannel::~OutputChannel()
 {
        if (has_current_frame) {
@@ -544,28 +550,24 @@ void Mixer::process_audio_one_frame()
        // Basically, if it's over about -40 dBFS, we squeeze it down to that level
        // (or more precisely, near it, since we don't use infinite ratio),
-       // then apply a makeup gain to get it to -12 dBFS. -12 dBFS is, of course,
+       // then apply a makeup gain to get it to -14 dBFS. -14 dBFS is, of course,
        // entirely arbitrary, but from practical tests with speech, it seems to
        // put ut around -23 LUFS, so it's a reasonable starting point for later use.